@@ -1,6 +1,4 @@
-# Copyright (c) Microsoft Corporation. All rights reserved.
-# Licensed under the MIT License.
-"""Implementation of tool support over LSP."""
+"""Implementation of the LSP server for Ruff."""
 
 from __future__ import annotations
 
@@ -51,32 +49,27 @@ from lsprotocol.types import (
 from pygls import protocol, server, uris, workspace
 from typing_extensions import TypedDict
 
-from ruff_lsp import utils
+from ruff_lsp import __version__, utils
 
+USER_DEFAULTS: dict[str, str] = {}
 WORKSPACE_SETTINGS: dict[str, dict[str, Any]] = {}
 INTERPRETER_PATHS: dict[str, str] = {}
 
 MAX_WORKERS = 5
 LSP_SERVER = server.LanguageServer(
     name="Ruff",
-    version="2022.0.23",
+    version=__version__,
     max_workers=MAX_WORKERS,
 )
 
-
-# **********************************************************
-# Tool specific code goes below this.
-# **********************************************************
-
 TOOL_MODULE = "ruff"
-
 TOOL_DISPLAY = "Ruff"
-
 TOOL_ARGS = ["--no-cache", "--no-fix", "--quiet", "--format", "json", "-"]
 
-# **********************************************************
-# Linting features start here
-# **********************************************************
+
+###
+# Linting.
+###
 
 
 @LSP_SERVER.feature(TEXT_DOCUMENT_DID_OPEN)
@@ -222,13 +215,9 @@ def hover(params: HoverParams) -> Hover | None:
     return None
 
 
-# **********************************************************
-# Linting features end here
-# **********************************************************
-
-# **********************************************************
-# Code Action features start here
-# **********************************************************
+###
+# Code Actions.
+###
 
 
 class TextDocument(TypedDict):
@@ -489,18 +478,30 @@ def _match_line_endings(document: workspace.Document, text: str) -> str:
     return text.replace(actual, expected)
 
 
-# **********************************************************
-# Code Action features ends here
-# **********************************************************
+###
+# Lifecycle.
+###
 
 
-# **********************************************************
-# Required Language Server Initialization and Exit handlers.
-# **********************************************************
 @LSP_SERVER.feature(INITIALIZE)
 def initialize(params: InitializeParams) -> None:
     """LSP handler for initialize request."""
-    settings = params.initialization_options["settings"]  # type: ignore
+    # Extract `settings` from the initialization options.
+    user_settings = params.initialization_options.get(  # type: ignore[attr-defined]
+        "settings",
+    )
+    if isinstance(user_settings, dict):
+        # In Sublime Text, Neovim, and probably others, we're passed a single
+        # `settings`, which we'll treat as defaults for any future files.
+        USER_DEFAULTS.update(user_settings)
+        settings = [user_settings]
+    elif isinstance(user_settings, list):
+        # In VS Code, we're passed a list of `settings`, one for each workspace folder.
+        # It doesn't really make sense to save these defaults.
+        settings = user_settings
+    else:
+        settings = []
+
     _update_workspace_settings(settings)
 
     if isinstance(LSP_SERVER.lsp, protocol.LanguageServerProtocol):
@@ -514,34 +515,49 @@ def initialize(params: InitializeParams) -> None:
             LSP_SERVER.lsp.trace = TraceValues.Off
 
 
-# *****************************************************
-# Internal functional and settings management APIs.
-# *****************************************************
-def _get_default_settings(workspace_path: str) -> dict[str, Any]:
+###
+# Settings.
+###
+
+
+def _default_settings() -> dict[str, Any]:
     return {
-        "check": False,
-        "workspaceFS": workspace_path,
-        "workspace": uris.from_fs_path(workspace_path),
-        "logLevel": "error",
-        "args": [],
-        "path": [],
-        "interpreter": [sys.executable],
-        "showNotifications": "off",
+        "logLevel": USER_DEFAULTS.get("logLevel", "error"),
+        "args": USER_DEFAULTS.get("args", []),
+        "path": USER_DEFAULTS.get("path", []),
+        "interpreter": USER_DEFAULTS.get("interpreter", [sys.executable]),
+        "importStrategy": USER_DEFAULTS.get("importStrategy", "fromEnvironment"),
+        "showNotifications": USER_DEFAULTS.get("showNotifications", "off"),
     }
 
 
-def _update_workspace_settings(settings) -> None:
+def _update_workspace_settings(settings: list[dict[str, Any]]) -> None:
     if not settings:
-        key = os.getcwd()
-        WORKSPACE_SETTINGS[key] = _get_default_settings(key)
+        workspace_path = os.getcwd()
+        WORKSPACE_SETTINGS[workspace_path] = {
+            **_default_settings(),
+            "workspaceFS": workspace_path,
+            "workspace": uris.from_fs_path(workspace_path),
+        }
         return
 
     for setting in settings:
-        key = uris.to_fs_path(setting["workspace"])
-        WORKSPACE_SETTINGS[key] = {
-            **setting,
-            "workspaceFS": key,
-        }
+        if "workspace" in setting:
+            workspace_path = uris.to_fs_path(setting["workspace"])
+            WORKSPACE_SETTINGS[workspace_path] = {
+                **_default_settings(),
+                **setting,
+                "workspaceFS": workspace_path,
+                "workspace": setting["workspace"],
+            }
+        else:
+            workspace_path = os.getcwd()
+            WORKSPACE_SETTINGS[workspace_path] = {
+                **_default_settings(),
+                **setting,
+                "workspaceFS": workspace_path,
+                "workspace": uris.from_fs_path(workspace_path),
+            }
 
 
 def _get_document_key(document: workspace.Document) -> str | None:
@@ -561,35 +577,68 @@ def _get_settings_by_document(document: workspace.Document | None) -> dict[str,
 
     key = _get_document_key(document)
     if key is None:
-        key = os.fspath(pathlib.Path(document.path).parent)
-        return _get_default_settings(key)
+        workspace_path = os.fspath(pathlib.Path(document.path).parent)
+        return {
+            **_default_settings(),
+            "workspaceFS": workspace_path,
+            "workspace": uris.from_fs_path(workspace_path),
+        }
 
     return WORKSPACE_SETTINGS[str(key)]
 
 
-# *****************************************************
+###
 # Internal execution APIs.
-# *****************************************************
-def _executable_path(settings: dict[str, Any]) -> list[str]:
+###
+
+
+def _executable_path(settings: dict[str, Any]) -> str:
     """Returns the path to the executable."""
+    bundle = get_bundle()
     if settings["path"]:
         # 'path' setting takes priority over everything.
-        return settings["path"]
-    elif settings["interpreter"] and not utils.is_current_interpreter(
+        log_to_output(f"Using 'path' setting: {settings['path']}")
+        for path in settings["path"]:
+            if os.path.exists(path):
+                return path
+
+    if settings["importStrategy"] == "useBundled" and bundle:
+        # If we're loading from the bundle, use the absolute path.
+        log_to_output(f"Using bundled executable: {bundle}")
+        return bundle
+
+    if settings["interpreter"] and not utils.is_current_interpreter(
         settings["interpreter"][0]
     ):
-        # If there is a different interpreter set, find its scripts path.
+        # If there is a different interpreter set, find its script path.
         if settings["interpreter"][0] not in INTERPRETER_PATHS:
             INTERPRETER_PATHS[settings["interpreter"][0]] = utils.scripts(
                 settings["interpreter"][0]
             )
-        return [
-            os.path.join(INTERPRETER_PATHS[settings["interpreter"][0]], TOOL_MODULE)
-        ]
+
+        path = os.path.join(INTERPRETER_PATHS[settings["interpreter"][0]], TOOL_MODULE)
+        if bundle and not os.path.exists(path):
+            log_to_output(
+                f"External interpreter executable ({path}) not found; "
+                f"falling back to bundled executable: {bundle}"
+            )
+            path = bundle
+        else:
+            log_to_output(f"Using external interpreter executable: {path}")
+        return path
+
+    # If the interpreter is same as the interpreter running this process, get the
+    # script path directly.
+    path = os.path.join(sysconfig.get_path("scripts"), TOOL_MODULE)
+    if bundle and not os.path.exists(path):
+        log_to_output(
+            f"Interpreter executable ({path}) not found; "
+            f"falling back to bundled executable: {bundle}"
+        )
+        path = bundle
     else:
-        # If the interpreter is same as the interpreter running this process, get the
-        # scripts path directly.
-        return [os.path.join(sysconfig.get_path("scripts"), TOOL_MODULE)]
+        log_to_output(f"Using interpreter executable: {path}")
+    return path
 
 
 def _run_tool_on_document(
@@ -613,7 +662,9 @@ def _run_tool_on_document(
     # Deep copy, to prevent accidentally updating global settings.
     settings = copy.deepcopy(_get_settings_by_document(document))
 
-    argv = _executable_path(settings) + TOOL_ARGS + settings["args"] + list(extra_args)
+    argv: list[str] = (
+        [_executable_path(settings)] + TOOL_ARGS + settings["args"] + list(extra_args)
+    )
     if use_stdin:
         argv += ["--stdin-filename", document.path]
     else:
@@ -640,7 +691,7 @@ def _run_subcommand_on_document(
     # Deep copy, to prevent accidentally updating global settings.
     settings = copy.deepcopy(_get_settings_by_document(document))
 
-    argv = _executable_path(settings) + list(args)
+    argv: list[str] = [_executable_path(settings)] + list(args)
 
     log_to_output(f"Running Ruff with: {argv}")
     result: utils.RunResult = utils.run_path(
@@ -654,9 +705,11 @@ def _run_subcommand_on_document(
     return result
 
 
-# *****************************************************
-# Logging and notification.
-# *****************************************************
+###
+# Logging.
+###
+
+
 def log_to_output(message: str, msg_type: MessageType = MessageType.Log) -> None:
     LSP_SERVER.show_message_log(message, msg_type)
 
@@ -683,8 +736,32 @@ def log_always(message: str) -> None:
         LSP_SERVER.show_message(message, MessageType.Info)
 
 
-# *****************************************************
-# Start the server.
-# *****************************************************
-if __name__ == "__main__":
+###
+# Bundled mode.
+###
+
+_BUNDLED_PATH: str | None = None
+
+
+def set_bundle(path: str) -> None:
+    """Sets the path to the bundled Ruff executable."""
+    global _BUNDLED_PATH
+    _BUNDLED_PATH = path
+
+
+def get_bundle() -> str | None:
+    """Returns the path to the bundled Ruff executable."""
+    return _BUNDLED_PATH
+
+
+###
+# Start up.
+###
+
+
+def start() -> None:
     LSP_SERVER.start_io()
+
+
+if __name__ == "__main__":
+    start()
